#pragma once

#include <map>
#include <mutex>
#include <optional>
#include <string>

#include "../filter/frequencyFilter/FrequencyFilter.h"
#include "../filter/equalityFilter/EqualityFilter.h"
#include "../extractor/imageExtractor/ImageExtractor.h"
#include "../converter/object/json/JsonConverter.h"
#include "../converter/image/png/PngConverter.h"

#include <RobotAPI/libraries/armem/core/MemoryID.h>

namespace armarx::armem::server::ltm
{
    /// all necessary classes to filter and convert an entry of the ltm to some other format(s)
    struct Processors
    {
        // Unique Memory Filters
        std::vector<MemoryFilter*> memFilters;
        filter::MemoryFrequencyFilter memFreqFilter;

        // Unique Snapshot filters
        std::vector<SnapshotFilter*> snapFilters;
        filter::SnapshotFrequencyFilter snapFreqFilter;
        filter::SnapshotEqualityFilter snapEqFilter;

        // Extractors
        std::vector<Extractor*> extractors;
        extractor::ImageExtractor imageExtractor;

        // Converters
        std::map<aron::type::Descriptor, Converter*> converters;
        converter::object::JsonConverter jsonConverter;
        converter::image::PngConverter pngConverter;

        Processors()
        {
            // setup containers
            memFilters.push_back(&memFreqFilter);
            snapFilters.push_back(&snapFreqFilter);
            snapFilters.push_back(&snapEqFilter);
            extractors.push_back(&imageExtractor);
            converters.insert({jsonConverter.convertsType, &jsonConverter});
            converters.insert({pngConverter.convertsType, &pngConverter});
        }
    };

    /// @brief Interface functions for the longterm memory classes
    class MemoryItem
    {
    public:
        MemoryItem(const MemoryID&); // only used by memory
        MemoryItem(const MemoryID&, const std::shared_ptr<Processors>&); // used by all other segments
        virtual ~MemoryItem() = default;

        MemoryID id() const;
        std::string name() const;

        virtual void setMemoryID(const MemoryID&);

    protected:
        std::shared_ptr<Processors> processors;

    private:
        MemoryID _id;
    };
} // namespace armarx::armem::server::ltm
