#include "CoreSegment.h"

#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>


namespace armarx::armem::server::ltm::disk
{
    CoreSegment::CoreSegment(const std::filesystem::path& p, const MemoryID& parentID, const std::string& segmentName, const std::shared_ptr<Processors>& filters) :
        CoreSegmentBase(parentID.withCoreSegmentName(segmentName), filters),
        DiskMemoryItem(p, parentID.memoryName, (EscapeSegmentName(segmentName) + "/"))
    {
    }

    bool CoreSegment::forEachProviderSegment(std::function<void(ProviderSegment&)>&& func) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return false;
        }

        //for (const auto& subdir : std::filesystem::directory_iterator(csPath))
        //{
        //    std::filesystem::path subdirPath = subdir.path();
        //    ProviderSegment c(csPath, id(), UnescapeSegmentName(subdirPath.filename()), processors);
        //    func(c);
        //}
        return true;
    }

    std::shared_ptr<ProviderSegment> CoreSegment::findProviderSegment(const std::string& providerSegmentName) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return nullptr;
        }

        std::filesystem::path relProvPath = relativeSegmentPath / (EscapeSegmentName(providerSegmentName) + "/");
        if (!util::checkIfFolderExists(mPath, relProvPath))
        {
            return nullptr;
        }

        auto c = std::make_shared<ProviderSegment>(memoryParentPath, id(), providerSegmentName, processors);
        return c;
    }

    void CoreSegment::_loadAllReferences(armem::wm::CoreSegment& e)
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        e.id() = id();

        forEachProviderSegment([&e](auto& x) {
            armem::wm::ProviderSegment s;
            x.loadAllReferences(s);
            e.addProviderSegment(s);
        });
    }

    void CoreSegment::_resolve(armem::wm::CoreSegment& c)
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        c.forEachProviderSegment([&](auto& e)
        {
            if (util::checkIfFolderExists(mPath, relativeSegmentPath))
            {
                ProviderSegment c(memoryParentPath, id(), e.id().providerSegmentName, processors);
                c.resolve(e);
            }
            else
            {
                ARMARX_WARNING << "Could not find the provider segment folder for segment '" << e.id().str() << "'.";
            }
        });
    }

    void CoreSegment::_store(const armem::wm::CoreSegment& c)
    {
        if (id().coreSegmentName.empty())
        {
            ARMARX_WARNING << "During storage of segment '" << c.id().str() << "' I noticed that the corresponding LTM has no id set. " <<
                              "I set the id of the LTM to the same name, however this should not happen!";
            id().coreSegmentName = c.id().coreSegmentName;
        }

        auto mPath = getMemoryBasePath();
        if (!util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            ARMARX_WARNING << "The segment folder for segment '"+id().str()+"'was not created. I will create the folder by myself, however it seems like there is a bug in the ltm pipeline.";
            util::ensureFolderExists(mPath, relativeSegmentPath, true);
        }

        c.forEachProviderSegment([&](const auto& prov)
        {
            std::filesystem::path relProvPath = relativeSegmentPath / (EscapeSegmentName(prov.id().providerSegmentName) + "/");
            util::ensureFolderExists(mPath, relProvPath, true);

            ProviderSegment c(memoryParentPath, id(), prov.id().providerSegmentName, processors);
            c.store(prov);
        });
    }

}
