// Header
#include "Entity.h"

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>

#include "../base/filter/frequencyFilter/FrequencyFilter.h"


namespace armarx::armem::server::ltm::disk
{
    Entity::Entity(const std::filesystem::path& p, const MemoryID& parentID, const std::string& segmentName, const std::shared_ptr<Processors>& filters) :
        EntityBase(parentID.withEntityName(UnescapeSegmentName(segmentName)), filters),
        DiskMemoryItem(p, parentID.memoryName, std::filesystem::path(EscapeSegmentName(parentID.coreSegmentName)) / EscapeSegmentName(parentID.providerSegmentName) / (EscapeSegmentName(segmentName) + "/"))
    {
    }

    bool Entity::forEachSnapshot(std::function<void(EntitySnapshot&)>&& func) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return false;
        }

        /*for (const auto& subdir : std::filesystem::directory_iterator(ePath))
        {
            std::filesystem::path subdirPath = subdir.path();
            if (!util::isNumber(subdirPath.filename()))
            {
                ARMARX_WARNING << "Found a non-timestamp folder inside an entity '" << id().str() << "' with name '" << subdirPath.filename() << "'. " <<
                                  "Ignoring this folder, however this is a bad situation.";
                continue;
            }

            EntitySnapshot c(ePath, id(), Time::microSeconds(std::stol(subdirPath.filename())), processors);
            func(c);
        }*/
        return true;
    }

    bool Entity::forEachSnapshotInIndexRange(long first, long last, std::function<void(EntitySnapshot&)>&& func) const
    {
        throw LocalException("NOT IMPLEMENTED YET BECAUSE THE DIRECTORY ITERATOR IS UNSORTED!");
    }

    bool Entity::forEachSnapshotInTimeRange(const Time& min, const Time& max, std::function<void(EntitySnapshot&)>&& func) const
    {
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts >= min && ts <= max)
            {
                func(e);
            }
        };

        return forEachSnapshot(std::move(f));
    }

    bool Entity::forEachSnapshotBeforeOrAt(const Time& time, std::function<void(EntitySnapshot&)>&& func) const
    {
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts <= time)
            {
                func(e);
            }
        };

        return forEachSnapshot(std::move(f));
    }

    bool Entity::forEachSnapshotBefore(const Time& time, std::function<void(EntitySnapshot&)>&& func) const
    {
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts < time)
            {
                func(e);
            }
        };

        return forEachSnapshot(std::move(f));
    }

    std::shared_ptr<EntitySnapshot> Entity::findSnapshot(const Time& n) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return nullptr;
        }

        std::filesystem::path tsPath = relativeSegmentPath / (std::to_string(n.toMicroSeconds()) + "/");
        if (!util::checkIfFolderExists(mPath, tsPath))
        {
            return nullptr;
        }

        return std::make_shared<EntitySnapshot>(memoryParentPath, id(), n, processors);
    }

    std::shared_ptr<EntitySnapshot> Entity::findLatestSnapshot() const
    {
        Time bestMatch = Time::microSeconds(-1);
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts > bestMatch)
            {
                bestMatch = ts;
            }
        };

        forEachSnapshot(std::move(f));

        if (bestMatch == Time::microSeconds(-1))
        {
            return nullptr;
        }

        return std::make_shared<EntitySnapshot>(memoryParentPath, id(), bestMatch, processors);
    }

    std::shared_ptr<EntitySnapshot> Entity::findLatestSnapshotBefore(const Time& time) const
    {
        Time bestMatch = Time::microSeconds(-1);
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts < time && ts > bestMatch)
            {
                bestMatch = ts;
            }
        };

        forEachSnapshot(std::move(f));

        if (bestMatch == Time::microSeconds(-1))
        {
            return nullptr;
        }

        return std::make_shared<EntitySnapshot>(memoryParentPath, id(), bestMatch, processors);
    }

    std::shared_ptr<EntitySnapshot> Entity::findLatestSnapshotBeforeOrAt(const Time& time) const
    {
        Time bestMatch = Time::microSeconds(-1);
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts <= time && ts > bestMatch)
            {
                bestMatch = ts;
            }
        };

        forEachSnapshot(std::move(f));

        if (bestMatch == Time::microSeconds(-1))
        {
            return nullptr;
        }

        return std::make_shared<EntitySnapshot>(memoryParentPath, id(), bestMatch, processors);
    }

    std::shared_ptr<EntitySnapshot> Entity::findFirstSnapshotAfter(const Time& time) const
    {
        Time bestMatch = Time::microSeconds(std::numeric_limits<long>::max());
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts > time && ts < bestMatch)
            {
                bestMatch = ts;
            }
        };

        forEachSnapshot(std::move(f));

        if (bestMatch == Time::microSeconds(std::numeric_limits<long>::max()))
        {
            return nullptr;
        }

        return std::make_shared<EntitySnapshot>(memoryParentPath, id(), bestMatch, processors);
    }

    std::shared_ptr<EntitySnapshot> Entity::findFirstSnapshotAfterOrAt(const Time& time) const
    {
        Time bestMatch = Time::microSeconds(std::numeric_limits<long>::max());
        auto f = [&](EntitySnapshot& e) {
            auto ts = e.id().timestamp;
            if (ts >= time && ts < bestMatch)
            {
                bestMatch = ts;
            }
        };

        forEachSnapshot(std::move(f));

        if (bestMatch == Time::microSeconds(std::numeric_limits<long>::max()))
        {
            return nullptr;
        }

        return std::make_shared<EntitySnapshot>(memoryParentPath, id(), bestMatch, processors);
    }

    void Entity::_loadAllReferences(armem::wm::Entity& e)
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        e.id() = id();

        forEachSnapshot([&e](auto& x) {
            if (!e.hasSnapshot(x.id().timestamp)) // we only load the references if the snapshot is not existant
            {
                armem::wm::EntitySnapshot s;
                x.loadAllReferences(s);
                e.addSnapshot(s);
            }
        });
    }

    void Entity::_resolve(armem::wm::Entity& p)
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath))
        {
            return;
        }

        p.forEachSnapshot([&](auto& e)
        {
            std::filesystem::path tsPath = relativeSegmentPath / (e.id().timestampStr() + "/");
            if (util::checkIfFolderExists(mPath, tsPath))
            {
                EntitySnapshot c(memoryParentPath, id(), e.id().timestamp, processors);
                c.resolve(e);
            }
            else
            {
                ARMARX_WARNING << "Could not find the snapshot segment folder for segment '" << e.id().str() << "'.";
            }
        });
    }

    void Entity::_store(const armem::wm::Entity& c)
    {
        if (id().entityName.empty())
        {
            ARMARX_WARNING << "During storage of segment '" << c.id().str() << "' I noticed that the corresponding LTM has no id set. " <<
                              "I set the id of the LTM to the same name, however this should not happen!";
            id().entityName = c.id().entityName;
        }

        auto mPath = getMemoryBasePath();
        if (!util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            ARMARX_WARNING << "The segment folder for segment '"+id().str()+"'was not created. I will create the folder by myself, however it seems like there is a bug in the ltm pipeline.";
            util::ensureFolderExists(mPath, relativeSegmentPath, true);
        }

        c.forEachSnapshot([&](const auto& snap)
        {
            std::filesystem::path tsPath = relativeSegmentPath / (snap.id().timestampStr() + "/");
            if (util::checkIfFolderExists(mPath, tsPath))
            {
                ARMARX_INFO << deactivateSpam() << "Ignoring to put an EntitiySnapshot into the LTM because the timestamp already existed (we assume snapshots are const and do not change outside the ltm).";
                return;
            }

            for (auto& f : processors->snapFilters)
            {
                if (f->enabled && !f->accept(snap))
                {
                    ARMARX_INFO << deactivateSpam() << "Ignoring to put an EntitiySnapshot into the LTM because it got filtered.";
                    return;
                }
            }

            util::ensureFolderExists(mPath, tsPath);
            EntitySnapshot c(memoryParentPath, id(), snap.id().timestamp, processors);
            c.store(snap);
        });
    }
}
