// Header
#include "EntitySnapshot.h"

// STD / STL
#include <iostream>
#include <fstream>

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <RobotAPI/libraries/aron/core/data/variant/container/Dict.h>

namespace armarx::armem::server::ltm::disk
{
    EntitySnapshot::EntitySnapshot(const std::filesystem::path& p, const MemoryID& parentID, const Time& ts, const std::shared_ptr<Processors>& filters) :
        EntitySnapshotBase(parentID.withTimestamp(ts), filters),
        DiskMemoryItem(p, parentID.memoryName, std::filesystem::path(EscapeSegmentName(parentID.coreSegmentName)) / EscapeSegmentName(parentID.providerSegmentName) / EscapeSegmentName(parentID.entityName) / (std::to_string(ts.toMicroSeconds()) + "/"))
    {
    }

    void EntitySnapshot::_loadAllReferences(armem::wm::EntitySnapshot& e) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        e.id() = id();

        for (unsigned int i = 0; i < 1000; ++i) // 1000 is max size for instances in a single timestamp
        {
            std::filesystem::path relIPath = relativeSegmentPath / (std::to_string(i) + "/");
            if (!util::checkIfFolderExists(mPath, relIPath))
            {
                break;
            }

            // add instance. Do not set data, since we only return references
            e.addInstance();
        }
    }

    void EntitySnapshot::_resolve(armem::wm::EntitySnapshot& e) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        auto& dictConverter = processors->converters.at(aron::type::Descriptor::eObject);

        // Get data from disk
        for (unsigned int i = 0; i < e.size(); ++i)
        {
            std::filesystem::path relIPath = relativeSegmentPath / (std::to_string(i) + "/");
            if (util::checkIfFolderExists(mPath, relIPath))
            {
                std::filesystem::path relDataPath = relIPath / (constantes::DATA_FILENAME + dictConverter->suffix);
                std::filesystem::path relMetadataPath = relIPath / (constantes::METADATA_FILENAME + dictConverter->suffix);

                auto& ins = e.getInstance(i);
                aron::data::DictPtr datadict = nullptr;
                aron::data::DictPtr metadatadict = nullptr;

                if (util::checkIfFileExists(mPath, relDataPath))
                {
                    auto datafilecontent = util::readDataFromFile(mPath, relDataPath);
                    auto dataaron = dictConverter->convert(datafilecontent);
                    datadict = aron::data::Dict::DynamicCastAndCheck(dataaron);

                    // check for special members
                    for (const auto& [key, m] : datadict->getElements())
                    {
                        for (auto& [t, f] : processors->converters)
                        {
                            if (t == aron::type::Descriptor::eObject)
                            {
                                continue;
                            }

                            std::filesystem::path relMemberPath = relIPath / (key + f->suffix);

                            if (util::checkIfFileExists(mPath, relMemberPath))
                            {
                                auto memberfilecontent = util::readDataFromFile(mPath, relMemberPath);
                                auto memberaron = f->convert(memberfilecontent);
                                datadict->setElement(key, memberaron);
                            }
                        }
                    }
                }
                else
                {
                    ARMARX_ERROR << "Could not find the data file '" << relDataPath.string() << "'. Continuing without data.";
                }

                if (util::checkIfFileExists(mPath, relMetadataPath))
                {
                    auto metadatafilecontent = util::readDataFromFile(mPath, relMetadataPath);
                    auto metadataaron = dictConverter->convert(metadatafilecontent);
                    metadatadict = aron::data::Dict::DynamicCastAndCheck(metadataaron);
                }
                else
                {
                    ARMARX_ERROR << "Could not find the metadata file '" << relMetadataPath.string() << "'. Continuing without metadata.";
                }

                from_aron(metadatadict, datadict, ins);
            }
            else
            {
                ARMARX_WARNING << "Could not find the index segment folder for segment '" << e.id().str() << "'.";
            }
        }
    }

    void EntitySnapshot::_store(const armem::wm::EntitySnapshot& e) const
    {
        if (id().timestampStr().empty())
        {
            ARMARX_WARNING << "During storage of segment '" << e.id().str() << "' I noticed that the corresponding LTM has no id set. " <<
                              "I set the id of the LTM to the same name, however this should not happen!";
            id().timestamp = e.id().timestamp;
        }

        auto& dictConverter = processors->converters.at(aron::type::Descriptor::eObject);

        auto mPath = getMemoryBasePath();
        if (!util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            ARMARX_WARNING << "The segment folder for segment '"+id().str()+"'was not created. I will create the folder by myself, however it seems like there is a bug in the ltm pipeline.";
            util::ensureFolderExists(mPath, relativeSegmentPath, true);
        }

        for (unsigned int i = 0; i < e.size(); ++i)
        {
            std::filesystem::path relIPath = relativeSegmentPath / (std::to_string(i) + "/");
            if (!util::checkIfFolderExists(mPath, relIPath))
            {
                util::ensureFolderExists(mPath, relIPath);

                std::filesystem::path relDataPath = relIPath / (constantes::DATA_FILENAME + dictConverter->suffix);
                std::filesystem::path relMetadataPath = relIPath / (constantes::METADATA_FILENAME + dictConverter->suffix);

                auto& ins = e.getInstance(i);

                // data
                auto dataAron = std::make_shared<aron::data::Dict>();
                auto metadataAron = std::make_shared<aron::data::Dict>();
                to_aron(metadataAron, dataAron, ins);

                // check special members for special extractions
                for (auto& x : processors->extractors)
                {
                    if (!x->enabled) continue;

                    auto t = x->extractsType;

                    Converter* conv = nullptr; // find suitable converter
                    for (const auto& [ct, c] : processors->converters)
                    {
                        if (!c->enabled) continue;
                        if (t != ct) continue;
                        conv = c;
                    }

                    if (conv)
                    {
                        auto dataExt = x->extract(dataAron);

                        for (const auto& [memberName, var] : dataExt.extraction)
                        {
                            ARMARX_CHECK_NOT_NULL(var);
                            std::filesystem::path relMemberPath = relIPath / (memberName + conv->suffix);
                            auto memberDataVec = conv->convert(var);
                            util::writeDataToFile(mPath, relMemberPath, memberDataVec);
                        }

                        dataAron = dataExt.dataWithoutExtraction;
                    }
                    // else we could not convert the extracted data so it makes no sense to extract it at all...
                }

                // convert dict and metadata
                auto dataVec = dictConverter->convert(dataAron);
                auto metadataVec = dictConverter->convert(metadataAron);
                util::writeDataToFile(mPath, relDataPath, dataVec);
                util::writeDataToFile(mPath, relMetadataPath, metadataVec);
            }
            // Ignore if the full index already exists
        }
    }
}
