#include "Memory.h"

#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>

namespace armarx::armem::server::ltm::disk
{
    void Memory::createPropertyDefinitions(PropertyDefinitionsPtr& properties, const std::string& prefix)
    {
        Base::createPropertyDefinitions(properties, prefix);
        properties->optional(memoryParentPath, prefix + "storagepath", "The path to the memory storage (the memory will be stored in a seperate subfolder).");
    }

    Memory::Memory() :
        Memory(std::filesystem::path("/tmp/MemoryExport"), "Test")
    {
    }

    Memory::Memory(const std::filesystem::path& p, const std::string& memoryName /* UNESCAPED */) :
        Base(MemoryID(memoryName, "")),
        DiskMemoryItem(p, memoryName, "")
    {
    }

    void Memory::setMemoryID(const MemoryID& id)
    {
        ARMARX_CHECK_NOT_EMPTY(id.memoryName);

        Base::setMemoryID(id.getMemoryID());

        memoryName = id.memoryName;
    }

    bool Memory::forEachCoreSegment(std::function<void(CoreSegment&)>&& func) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath))
        {
            return false;
        }

        //for (const auto& subdir : std::filesystem::directory_iterator(mPath))
        //{
        //    std::filesystem::path subdirPath = subdir.path();
        //    CoreSegment c(memoryParentPath, id(), UNESCPAE, processors);
        //    func(c);
        //}
        return true;
    }

    std::shared_ptr<CoreSegment> Memory::findCoreSegment(const std::string& coreSegmentName) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath))
        {
            return nullptr;
        }

        std::filesystem::path relCsPath = (EscapeSegmentName(coreSegmentName) + "/");
        if (!util::checkIfFolderExists(mPath, relCsPath))
        {
            return nullptr;
        }

        auto c = std::make_shared<CoreSegment>(memoryParentPath, id(), coreSegmentName, processors);
        return c;
    }

    void Memory::_loadAllReferences(armem::wm::Memory& m)
    {
        m.id() = id();

        forEachCoreSegment([&m](auto& x) {
            armem::wm::CoreSegment s;
            x.loadAllReferences(s);
            m.addCoreSegment(s);
        });
    }

    void Memory::_resolve(armem::wm::Memory& m)
    {
        std::lock_guard l(ltm_mutex); // we cannot load a memory multiple times simultaneously
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath))
        {
            return;
        }

        m.forEachCoreSegment([&](auto& e)
        {
            std::filesystem::path relCsPath = EscapeSegmentName(e.id().coreSegmentName) + "/";
            if (util::checkIfFolderExists(mPath, relCsPath))
            {
                CoreSegment c(memoryParentPath, id(), e.id().coreSegmentName, processors);
                c.resolve(e);
            }
            else
            {
                ARMARX_WARNING << "Could not find the core segment folder for segment '" << e.id().str() << "'.";
            }
        });
    }

    void Memory::_directlyStore(const armem::wm::Memory& memory)
    {
        std::lock_guard l(ltm_mutex); // we cannot store a memory multiple times simultaneously

        if (id().memoryName.empty())
        {
            ARMARX_WARNING << "During storage of memory '" << memory.id().str() << "' I noticed that the corresponding LTM has no id set. " <<
                              "I set the id of the LTM to the same name, however this should not happen!";
            setMemoryID(memory.id());
        }

        auto mPath = getMemoryBasePath();
        util::ensureBasePathExists(mPath, true); // create if not exists

        memory.forEachCoreSegment([&](const auto& core)
        {
            std::filesystem::path relCsPath = (EscapeSegmentName(core.id().coreSegmentName) + "/");
            util::ensureFolderExists(mPath, relCsPath, true); // create subfolder

            CoreSegment c(memoryParentPath, id(), core.id().coreSegmentName, processors);
            c.store(core);
        });
    }
}
