// Header
#include "ProviderSegment.h"

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>


namespace armarx::armem::server::ltm::disk
{
    ProviderSegment::ProviderSegment(const std::filesystem::path& p, const MemoryID& parentID, const std::string& segmentName, const std::shared_ptr<Processors>& filters) :
        ProviderSegmentBase(parentID.withProviderSegmentName(segmentName), filters),
        DiskMemoryItem(p, parentID.memoryName, std::filesystem::path(EscapeSegmentName(parentID.coreSegmentName)) / (EscapeSegmentName(segmentName) + "/"))
    {
    }

    bool ProviderSegment::forEachEntity(std::function<void(Entity&)>&& func) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return false;
        }

        //for (const auto& subdir : std::filesystem::directory_iterator(provPath))
        //{
        //    std::filesystem::path subdirPath = subdir.path();
        //    Entity c(provPath, id(), subdirPath.filename(), processors);
        //    func(c);
        //}
        return true;
    }

    std::shared_ptr<Entity> ProviderSegment::findEntity(const std::string& entityName) const
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return nullptr;
        }

        std::filesystem::path relEPath = relativeSegmentPath / (EscapeSegmentName(entityName) + "/");
        if (!util::checkIfFolderExists(mPath, relEPath))
        {
            return nullptr;
        }

        auto c = std::make_shared<Entity>(memoryParentPath, id(), entityName, processors);
        return c;
    }

    void ProviderSegment::_loadAllReferences(armem::wm::ProviderSegment& e)
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        e.id() = id();

        forEachEntity([&e](auto& x) {
            armem::wm::Entity s;
            x.loadAllReferences(s);
            e.addEntity(s);
        });
    }

    void ProviderSegment::_resolve(armem::wm::ProviderSegment& p)
    {
        auto mPath = getMemoryBasePath();
        if (!util::checkIfBasePathExists(mPath) || !util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            return;
        }

        p.forEachEntity([&](auto& e)
        {
            std::filesystem::path relEPath = relativeSegmentPath / (EscapeSegmentName(e.id().entityName) + "/");
            if (util::checkIfFolderExists(mPath, relEPath))
            {
                Entity c(memoryParentPath, id(), e.id().entityName, processors);
                c.resolve(e);
            }
            else
            {
                ARMARX_WARNING << "Could not find the entity segment folder for segment '" << e.id().str() << "'.";
            }
        });
    }

    void ProviderSegment::_store(const armem::wm::ProviderSegment& p)
    {
        if (id().providerSegmentName.empty())
        {
            ARMARX_WARNING << "During storage of segment '" << p.id().str() << "' I noticed that the corresponding LTM has no id set. " <<
                              "I set the id of the LTM to the same name, however this should not happen!";
            id().providerSegmentName = p.id().providerSegmentName;
        }

        auto mPath = getMemoryBasePath();
        if (!util::checkIfFolderExists(mPath, relativeSegmentPath))
        {
            ARMARX_WARNING << "The segment folder for segment '"+id().str()+"'was not created. I will create the folder by myself, however it seems like there is a bug in the ltm pipeline.";
            util::ensureFolderExists(mPath, relativeSegmentPath, true);
        }

        p.forEachEntity([&](const auto& e)
        {
            std::filesystem::path relEPath = relativeSegmentPath / (EscapeSegmentName(e.id().entityName) + "/");
            util::ensureFolderExists(mPath, relEPath, true);

            Entity c(memoryParentPath, id(), e.id().entityName, processors);
            c.store(e);
        });
    }
}
