#pragma once

// STD / STL
#include <filesystem>
#include <iostream>
#include <fstream>

#include <minizip/zip.h>
#include <minizip/unzip.h>

#include "../../../../core/error.h"

namespace armarx::armem::server::ltm::disk
{
    namespace constantes
    {
        const std::string TYPE_FILENAME = "type.aron";
        const std::string DATA_FILENAME = "data.aron";
        const std::string METADATA_FILENAME = "metadata.aron";
    }

    namespace filesystem::util
    {
        inline bool checkIfFolderInFilesystemExists(const std::filesystem::path& p)
        {
            return std::filesystem::exists(p) and std::filesystem::is_directory(p);
        }

        inline void ensureFolderInFilesystemExists(const std::filesystem::path& p, bool createIfNotExistent = true)
        {
            if (!std::filesystem::exists(p))
            {
                if (createIfNotExistent)
                {
                    std::filesystem::create_directories(p);
                }
            }
            if (!std::filesystem::is_directory(p))
            {
                throw error::ArMemError("Could not find folder: " + p.string());
            }
        }

        inline bool checkIfFileInFilesystemExists(const std::filesystem::path& p)
        {
            return std::filesystem::exists(p) and std::filesystem::is_regular_file(p);
        }

        inline void ensureFileInFilesystemExists(const std::filesystem::path& p)
        {
            if (!std::filesystem::exists(p) || !std::filesystem::is_regular_file(p))
            {
                // not found
                throw error::ArMemError("Could not find file: " + p.string());
            }
        }

        inline void writeDataInFilesystemFile(const std::filesystem::path& p, const std::vector<unsigned char>& data)
        {
            std::ofstream dataofs;
            dataofs.open(p);
            if (!dataofs)
            {
                throw error::ArMemError("Could not write data to filesystem file '" + p.string() + "'. Skipping this file.");
            }
            dataofs.write(reinterpret_cast<const char*>(data.data()), data.size());
            dataofs.close();
        }

        inline std::vector<unsigned char> readDataFromFilesystemFile(const std::filesystem::path path)
        {
            std::ifstream dataifs(path);
            std::vector<unsigned char> datafilecontent((std::istreambuf_iterator<char>(dataifs)), (std::istreambuf_iterator<char>()));
            return datafilecontent;
        }
    }

    namespace minizip::util
    {
        const std::string MINIZIP_SUFFIX = ".zip";
        const int MINIZIP_CASE_SENSITIVITY = 1;

        inline bool checkZipFile(const std::filesystem::path& pathToZip)
        {
            bool ret = false;
            zipFile z = NULL;
            if (std::filesystem::exists(pathToZip))
            {
                if (std::filesystem::is_regular_file(pathToZip) && pathToZip.extension() == MINIZIP_SUFFIX)
                {
                    z = zipOpen(pathToZip.string().c_str(), APPEND_STATUS_ADDINZIP);
                    ret = (bool) z;
                    zipClose(z, NULL);
                }
            }
            return ret;
        }

        inline zipFile ensureZipFile(const std::filesystem::path& pathToZip, bool createIfNotExistent = true)
        {
            int mode = APPEND_STATUS_CREATE;
            if (!checkZipFile(pathToZip))
            {
                if (createIfNotExistent)
                {
                    mode = APPEND_STATUS_CREATE;
                }
                else
                {
                    throw error::ArMemError("Could not find zip file: " + pathToZip.string());
                }
            }

            zipFile z = zipOpen(pathToZip.string().c_str(), mode);
            if (!z)
            {
                throw error::ArMemError("Unknown error occured during opening zip file: " + pathToZip.string());
            }
            return z;
        }

        inline unzFile getUnzFile(const std::filesystem::path& pathToZip)
        {
            unzFile z = NULL;
            if (std::filesystem::exists(pathToZip))
            {
                if (std::filesystem::is_regular_file(pathToZip) && pathToZip.extension() == MINIZIP_SUFFIX)
                {
                    z = unzOpen(pathToZip.string().c_str());
                }
                else
                {
                    throw error::ArMemError("Could not replace existing file: " + pathToZip.string());
                }
            }

            if (!z)
            {
                throw error::ArMemError("Unknown error occured during opening unz file: " + pathToZip.string());
            }
            return z;
        }

        inline bool checkIfElementInZipExists(const std::filesystem::path& pathToZip, const std::filesystem::path& p)
        {
            if (!checkZipFile(pathToZip)) return false;
            auto uzf = getUnzFile(pathToZip);

            bool ret = (unzLocateFile(uzf, p.string().c_str(), MINIZIP_CASE_SENSITIVITY) == UNZ_OK);

            unzClose(uzf);
            return ret;
        }

        inline void ensureElementInZipExists(const std::filesystem::path& pathToZip, const std::filesystem::path& p, bool createIfNotExistent)
        {
            auto zf = ensureZipFile(pathToZip);
            auto uzf = getUnzFile(pathToZip);

            if (auto r = unzLocateFile(uzf, p.string().c_str(), MINIZIP_CASE_SENSITIVITY); r != UNZ_OK)
            {
                if (createIfNotExistent)
                {
                    zip_fileinfo zfi;
                    zipOpenNewFileInZip(zf, p.string().c_str(), &zfi, NULL, 0, NULL, 0, NULL, Z_DEFLATED, Z_DEFAULT_COMPRESSION);
                    zipCloseFileInZip(zf);
                }
                else
                {
                    unzClose(uzf);
                    zipClose(zf, NULL);
                    throw error::ArMemError("Could not find element '" + p.string() + "' in zip file: " + pathToZip.string());
                }
            }
            // else folder exists

            unzClose(uzf);
            zipClose(zf, NULL);
        }

        inline void ensureFolderInZipExists(const std::filesystem::path& pathToZip, const std::filesystem::path& p, bool createIfNotExistent = true)
        {
            if (!p.filename().empty())
            {
                throw error::ArMemError("The specified path is not a folder (it needs tailing /): " + p.string());
            }

            ensureElementInZipExists(pathToZip, p, createIfNotExistent);
        }

        inline void ensureFileInZipExists(const std::filesystem::path& pathToZip, const std::filesystem::path& p)
        {
            if (p.filename().empty())
            {
                throw error::ArMemError("The specified path is not a file (it needs a filename): " + p.string());
            }

            ensureElementInZipExists(pathToZip, p, true);
        }

        inline bool checkIfFolderInZipExists(const std::filesystem::path& pathToZip, const std::filesystem::path& p)
        {
            if (!p.filename().empty())
            {
                throw error::ArMemError("The specified path is not a folder (it needs tailing /): " + p.string());
            }

            return checkIfElementInZipExists(pathToZip, p);
        }

        inline bool checkIfFileInZipExists(const std::filesystem::path& pathToZip, const std::filesystem::path& p)
        {
            if (p.filename().empty())
            {
                throw error::ArMemError("The specified path is not a file (it needs a filename): " + p.string());
            }

            return checkIfElementInZipExists(pathToZip, p);
        }

        inline void writeDataInFileInZipFile(const std::filesystem::path& pathToZip, const std::filesystem::path& p, const std::vector<unsigned char>& data)
        {
            auto zf = ensureZipFile(pathToZip);

            zip_fileinfo zfi;
            zipOpenNewFileInZip(zf, p.string().c_str(), &zfi, NULL, 0, NULL, 0, NULL, Z_DEFLATED, Z_DEFAULT_COMPRESSION);
            zipWriteInFileInZip(zf, reinterpret_cast<const char*>(data.data()), data.size());
            zipCloseFileInZip(zf);
            zipClose(zf, NULL);
        }

        inline std::vector<unsigned char> readDataFromFileInZipFile(const std::filesystem::path& pathToZip, const std::filesystem::path& p)
        {
            auto uzf = getUnzFile(pathToZip);
            if (unzLocateFile(uzf, p.string().c_str(), MINIZIP_CASE_SENSITIVITY) == UNZ_OK) // set current file
            {
                // File located
                unz_file_info uzfi;
                if (unzGetCurrentFileInfo(uzf, &uzfi, NULL, 0, NULL, 0, NULL, 0) == UNZ_OK)
                {
                    std::vector<unsigned char> data(uzfi.uncompressed_size);

                    if (unzOpenCurrentFile(uzf) == UNZ_OK) // open current file
                    {
                        if (unzReadCurrentFile(uzf, reinterpret_cast<char*>(data.data()), uzfi.uncompressed_size) == UNZ_OK) // read file in buffer
                        {
                            unzCloseCurrentFile(uzf);
                            unzClose(uzf);
                            return data;
                        }
                    }
                }
            }
            unzClose(uzf);
            throw error::ArMemError("Could not read data from zip file '" + pathToZip.string() + "' and internal path '" + p.string() + "'.");
        }
    }


    namespace util
    {
        // check whether a string is a number (e.g. timestamp folders)
        inline bool isNumber(const std::string& s)
        {
            for (char const& ch : s)
            {
                if (std::isdigit(ch) == 0)
                {
                    return false;
                }
            }
            return true;
        }

        inline bool checkIfBasePathExists(const std::filesystem::path& mPath)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::checkZipFile(mPath);
            }
            return filesystem::util::checkIfFolderInFilesystemExists(mPath);
        }

        inline void ensureBasePathExists(const std::filesystem::path& mPath, bool createIfNotExistent = true)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                minizip::util::ensureZipFile(mPath, createIfNotExistent);
                return;
            }
            return filesystem::util::ensureFolderInFilesystemExists(mPath, createIfNotExistent);
        }

        inline bool checkIfFolderExists(const std::filesystem::path& mPath, const std::filesystem::path& p)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::checkIfFolderInZipExists(mPath, p);
            }

            return filesystem::util::checkIfFolderInFilesystemExists(mPath / p);
        }

        inline void ensureFolderExists(const std::filesystem::path& mPath, const std::filesystem::path& p, bool createIfNotExistent = true)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::ensureFolderInZipExists(mPath, p, createIfNotExistent);
            }

            return filesystem::util::ensureFolderInFilesystemExists(mPath / p, createIfNotExistent);
        }

        inline bool checkIfFileExists(const std::filesystem::path& mPath, const std::filesystem::path& p)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::checkIfFileInZipExists(mPath, p);
            }

            return filesystem::util::checkIfFileInFilesystemExists(mPath / p);
        }

        inline void ensureFileExists(const std::filesystem::path& mPath, const std::filesystem::path& p)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::ensureFileInZipExists(mPath, p);
            }

            return filesystem::util::ensureFileInFilesystemExists(mPath / p);
        }

        inline void writeDataToFile(const std::filesystem::path& mPath, const std::filesystem::path& p, const std::vector<unsigned char>& data)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::writeDataInFileInZipFile(mPath, p, data);
            }

            return filesystem::util::writeDataInFilesystemFile(mPath / p, data);
        }

        inline std::vector<unsigned char> readDataFromFile(const std::filesystem::path& mPath, const std::filesystem::path& p)
        {
            if (mPath.extension() == minizip::util::MINIZIP_SUFFIX)
            {
                return minizip::util::readDataFromFileInZipFile(mPath, p);
            }

            return filesystem::util::readDataFromFilesystemFile(mPath / p);
        }
    }
} // namespace armarx::armem::server::ltm::disk
