#include "SkillImplementationWrapper.h"

namespace armarx
{
    namespace skills::detail
    {
        void SkillImplementationWrapper::execute(const skills::SkillParameterization parameterization)
        {
            std::unique_lock l(executingMutex);

            const std::string skillName = skill->description.skillName;
            ARMARX_INFO_S << "Executing skill: " << skillName;

            // reset Skill
            {
                std::unique_lock l2(skillStatusMutex); // skill is not updating
                statusUpdate.status = skills::provider::dto::Execution::Status::Idle;
                statusUpdate.data = nullptr;
                skill->reset();
            }

            // set params and setup variables
            {
                std::lock_guard l(skillStatusMutex);
                statusUpdate.usedParameterization = parameterization;
            }

            auto& aron_params = parameterization.usedInputParams;
            auto updateStatus = [&](const skills::provider::dto::Execution::Status status, const aron::data::DictPtr& data = nullptr){
                std::lock_guard l(skillStatusMutex);
                statusUpdate.status = status;
                statusUpdate.data = data;

                auto& callbackInterface = statusUpdate.usedParameterization.usedCallbackInterface;

                if (callbackInterface)
                {
                    callbackInterface->updateStatusForSkill(statusUpdate.toIce());
                }
            };

            try
            {
                // Check params
                if (not(aron_params) && skill->description.acceptedType)
                {
                    throw armarx::LocalException("The Skill '" + skillName + "' requires a type but no params are NULL.");
                }

                // set scheduled
                updateStatus(skills::provider::dto::Execution::Status::Scheduled);

                // execute
                updateStatus(skills::provider::dto::Execution::Status::Running);

                skill->init(aron_params);
                auto ret = skill->execute(aron_params, [&](const aron::data::DictPtr& update) { updateStatus(statusUpdate.status, update); });
                skill->exit(aron_params);

                switch (ret)
                {
                    case skills::Skill::Status::Failed:
                        throw armarx::LocalException("The Skill '" + skillName + "' failed during execution.");
                        break;
                    case skills::Skill::Status::TimeoutReached:
                        throw armarx::LocalException("The Skill '" + skillName + "' reached timeout during execution.");
                        break;
                    case skills::Skill::Status::Stopped:
                    {
                        updateStatus(skills::provider::dto::Execution::Status::Aborted);
                        break;
                    }
                    case skills::Skill::Status::Succeeded:
                    {
                        updateStatus(skills::provider::dto::Execution::Status::Succeeded);
                        break;
                    }
                }
            }
            catch (const std::exception& ex)
            {
                ARMARX_WARNING_S << "Skill " << skillName << " died with exception:\n" << ex.what();
                updateStatus(skills::provider::dto::Execution::Status::Failed);
            }
        }
    }
}
